# The MIT License (MIT)
#
# Copyright (c) 2015-2016 Massachusetts Institute of Technology.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import division

import datetime
import random
import math
import csv
import copy

from bandicoot.core import Record, Position
import bandicoot as bc
from bandicoot.helper.tools import OrderedDict


def _choice(seq):
    return seq[int(random.random() * len(seq))]


def _uniform(a, b):
    return a + (b - a) * random.random()


def _randint(a, b):
    return int(_uniform(a, b))


def _sample(population, k):
    n = len(population)
    result = [None] * k

    selected = set()
    selected_add = selected.add
    for i in range(k):
        j = _randint(0, n)
        while j in selected:
            j = _randint(0, n)
        selected_add(j)
        result[i] = population[j]
    return result


def random_record(**kwargs):
    n_users = 48
    rate = 1e-4

    year = 2012
    month = _choice(range(1, 3))
    day = _choice(range(1, 28))

    # ensures that some correspondents have more interactions than others
    correspondent = int(random.random() * n_users)

    r = {'datetime': datetime.datetime(year, month, day) + datetime.timedelta(seconds=math.floor(-1 / rate * math.log(random.random()))),
         'interaction': _choice(['text', 'text', 'text', 'call', 'call']),
         'correspondent_id': "correspondent_{}".format(correspondent),
         'direction': _choice(['in', 'in', 'out']),
         'call_duration': int(random.random() * 1000),
         'position': Position(location=(_uniform(-5, 5), _uniform(-5, 5)))}
    if r['interaction'] == "text":
        r['call_duration'] = None

    r.update(kwargs)
    return Record(**r)


def sample_user(number_records=1482, seed=42, pct_in_network=0.8):
    old_state = random.getstate()

    try:
        random.seed(seed, version=1)
    except TypeError:
        random.seed(seed)

    towers = {
        701: (42.3555, -71.099541),
        702: (42.359039, -71.094595),
        703: (42.360481, -71.087321),
        704: (42.361013, -71.097868),
        705: (42.370849, -71.114613),
        706: (42.3667427, -71.1069847),
        707: (42.367589, -71.076537)
    }
    towers_position = [Position(antenna=k, location=v)
                       for k, v in towers.items()]

    ego_records = [random_record(
        position=_choice(towers_position)) for _ in range(number_records)]
    user, _ = bc.io.load(
        "sample_user", ego_records, towers, None, describe=False)

    # create network
    correspondents = set([record.correspondent_id for record in ego_records])
    correspondents = sorted(list(correspondents))
    correspondent_records = {}
    connections = {}

    n_in_network = int(len(correspondents) * pct_in_network)
    n_in_network = (n_in_network // 2) * 2
    in_network_correspondents = _sample(correspondents, n_in_network)

    def reverse_records(records, current_owner):
        for r in records:
            r.direction = 'out' if r.direction == 'in' else 'in'
            r.correspondent_id = current_owner
        return records

    # set records from ego
    for c_id in sorted(in_network_correspondents):
        reciprocal_records = [
            r for r in ego_records if r.correspondent_id == c_id]
        reciprocal_records = reverse_records(
            copy.deepcopy(reciprocal_records), "sample_user")
        correspondent_records[c_id] = reciprocal_records

    def generate_group_with_random_links(pct_users_in_group):
        n_in_group = int(len(correspondents) * pct_users_in_group)
        group = _sample(non_grouped_correspondents, n_in_group)
        networkusers_group = list()
        for user in group:
            if user in in_network_correspondents:
                networkusers_group.append(user)

        def create_pair(source):
            user_pair = [source, _choice(group)]
            if user_pair[0] in non_grouped_correspondents:
                non_grouped_correspondents.remove(user_pair[0])
            if user_pair[1] in non_grouped_correspondents:
                non_grouped_correspondents.remove(user_pair[1])

            extra_records = [
                random_record(position=_choice(towers_position),
                              interaction=_choice(['text', 'call', 'call']),
                              correspondent_id=user_pair[1])
                for _ in range(_randint(25, 150))]

            correspondent_records[user_pair[0]].extend(extra_records)
            if (user_pair[1] in in_network_correspondents):
                rr = reverse_records(copy.deepcopy(extra_records), user_pair[0])
                correspondent_records[user_pair[1]].extend(rr)

        # create pairs of users
        for i in range(len(networkusers_group)):
            create_pair(networkusers_group[i])
            if random.random() > 0.5:
                create_pair(networkusers_group[i])

    non_grouped_correspondents = copy.deepcopy(correspondents)
    for i in range(4):
        generate_group_with_random_links(pct_users_in_group=0.4 - i * 0.1)

    # create user object
    for c_id in sorted(correspondents):
        if c_id in in_network_correspondents:
            correspondent_user, _ = bc.io.load(
                c_id, correspondent_records[c_id], towers, None, describe=False)
        else:
            correspondent_user = None
        connections[c_id] = correspondent_user

    # return the network dictionary sorted by key
    user.network = OrderedDict(sorted(connections.items(), key=lambda t: t[0]))
    user.recompute_missing_neighbors()

    random.setstate(old_state)
    return user


def write_new_user(filepath, n=1960):
    user = sample_user(n)

    schema = ['interaction', 'direction', 'correspondent_id',
              'datetime', 'call_duration', 'antenna_id']
    with open(filepath, "wb") as new_user:
        w = csv.writer(new_user)
        w.writerow(schema)

        for record in user.records:
            w.writerow([record.position.antenna if x ==
                        'antenna_id' else getattr(record, x) for x in schema])

    print("Finished writing new user to " + filepath)
    write_answers(user)


def write_answers(user):
    bc.io.to_json([bc.utils.all(user, groupby='week', summary='default')],
                  "samples/automatic/automatic_result.json")
    bc.io.to_json([bc.utils.all(user, groupby='week', summary='extended')],
                  "samples/automatic/automatic_result_extended.json")
    bc.io.to_json([bc.utils.all(user, groupby=None, summary='default')],
                  "samples/automatic/automatic_result_no_grouping.json")


def random_burst(count, delta=datetime.timedelta(minutes=10), **kwargs):

    first_date = datetime.datetime(2014, 1, 1, 10, 41)

    for i in range(count):
        _date = first_date + delta * i
        yield random_record(datetime=_date, **kwargs)



# The MIT License (MIT)
#
# Copyright (c) 2015-2016 Massachusetts Institute of Technology.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from bandicoot.helper.maths import great_circle_distance
import itertools


def compute_distance_matrix(points):
    """
    Return a matrix of distance (in meters) between every point in a given list
    of (lat, lon) location tuples.
    """
    n = len(points)
    return [[1000 * great_circle_distance(points[i], points[j])
             for j in range(n)] for i in range(n)]


def get_neighbors(distance_matrix, source, eps):
    """
    Given a matrix of distance between couples of points,
    return the list of every point closer than eps from a certain point.
    """

    return [dest for dest, distance in enumerate(distance_matrix[source]) if distance < eps]


def dbscan(points, eps, minpts):
    """
    Implementation of [DBSCAN]_ (*A density-based algorithm for discovering
    clusters in large spatial databases with noise*). It accepts a list of
    points (lat, lon) and returns the labels associated with the points.

    References
    ----------
    .. [DBSCAN] Ester, M., Kriegel, H. P., Sander, J., & Xu, X. (1996, August).
        A density-based algorithm for discovering clusters in large
        spatial databases with noise. In Kdd (Vol. 96, No. 34, pp. 226-231).

    """
    next_label = 0
    n = len(points)
    labels = [None] * n

    distance_matrix = compute_distance_matrix(points)
    neighbors = [get_neighbors(distance_matrix, i, eps) for i in range(n)]

    for i in range(n):
        if labels[i] is not None:
            continue

        if len(neighbors[i]) < minpts:
            continue

        labels[i] = next_label
        candidates = [i]
        while len(candidates) > 0:
            c = candidates.pop()

            for j in neighbors[c]:
                if labels[j] is None:
                    labels[j] = next_label
                    if len(neighbors[j]) >= minpts:
                        candidates.append(j)

        next_label += 1

    return labels


def _groupwhile(x, fwhile):
    groups = []
    i = 0
    while i < len(x):
        j = i
        while j < len(x) - 1 and fwhile(i, j + 1):
            j = j + 1
        groups.append(x[i:j + 1])
        i = j + 1
    return groups


def get_stops(records, group_dist):
    """
    Group records arounds stop locations and returns a list of
    dict(location, records) for each stop.

    Parameters
    ----------
    records : list
        A list of Record objects ordered by non-decreasing datetime
    group_dist : float
        Minimum distance (in meters) to switch to a new stop.
    """
    def traverse(start, next):
        position_prev = records[next - 1].position.location
        position_next = records[next].position.location
        dist = 1000 * great_circle_distance(position_prev, position_next)
        return dist <= group_dist

    groups = _groupwhile(records, traverse)

    def median(x):
        return sorted(x)[len(x) // 2]

    stops = []
    for g in groups:
        _lat = median([gv.position.location[0] for gv in g])
        _lon = median([gv.position.location[1] for gv in g])
        stops.append({
            'location': (_lat, _lon),
            'records': g,
        })

    return stops


def cluster_and_update(records, group_dist=50, eps=100):
    """
    Update records, by clustering their positions using the DBSCAN algorithm.
    Returns a dictionnary associating new antenna identifiers to (lat, lon)
    location tuples.

    .. note:: Use this function to cluster fine-grained GPS records.

    Parameters
    ----------
    records : list
        A list of Record objects ordered by non-decreasing datetime
    group_dist : float, default: 50
        Minimum distance (in meters) to switch to a new stop.
    eps : float, default: 100
        The eps parameter for the DBSCAN algorithm.
    """
    # Run the DBSCAN algorithm with all stop locations and 1 minimal point.
    stops = get_stops(records, group_dist)
    labels = dbscan([s['location'] for s in stops], eps, 1)

    # Associate all records with their new antenna
    antennas = {}
    for i, stop in enumerate(stops):
        antenna_id = labels[i]
        for record in stop['records']:
            record.position.antenna = antenna_id

        antennas[antenna_id] = stop['location']

    all_records = sum(len(stop['records']) for stop in stops)
    if all_records != len(records):
        raise ValueError("get_antennas has {} records instead of {} "
                         "initially.".format(all_records, len(records)))

    return antennas


def fix_location(records, max_elapsed_seconds=300):
    """
    Update position of all records based on the position of
    the closest GPS record.

    .. note:: Use this function when call and text records are missing a
              location, but you have access to accurate GPS traces.

    """
    groups = itertools.groupby(records, lambda r: r.direction)
    groups = [(interaction, list(g)) for interaction, g in groups]

    def tdist(t1, t2):
            return abs((t1 - t2).total_seconds())

    for i, (interaction, g) in enumerate(groups):
        if interaction == 'in':
            continue

        prev_gps = groups[i-1][1][-1]
        next_gps = groups[i+1][1][0]

        for r in g:
            if tdist(r.datetime, prev_gps.datetime) <= max_elapsed_seconds:
                r.position = prev_gps.position
            elif tdist(r.datetime, next_gps.datetime) <= max_elapsed_seconds:
                r.position = next_gps.position



# The MIT License (MIT)
#
# Copyright (c) 2015-2016 Massachusetts Institute of Technology.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Tests for bandicoot.core (User, Position, and Record classes)
"""

import bandicoot as bc
from .testing_tools import parse_dict

import unittest
import datetime
import sys
import os

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO


class TestCore(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._dir_changed = False

    def setUp(self):
        if not TestCore._dir_changed:
            abspath = os.path.abspath(__file__)
            name = abspath.index(os.path.basename(__file__))
            abspath = abspath[:name]
            os.chdir(abspath)
            TestCore._dir_changed = True

        self.user = bc.io.read_orange("u_test", "samples", describe=False)

        self.user_nocturnal = bc.io.read_orange(
            "u_test", "samples", describe=False)
        self.user_nocturnal.night_start = datetime.time(7, 0)
        self.user_nocturnal.night_end = datetime.time(19, 0)

        self.maxDiff = None

    def test_recompute_home_place(self):
        old_home = self.user.home
        new_home = self.user.recompute_home()

        self.assertEqual(old_home, new_home)
        self.assertEqual(old_home, self.user.home)

    def test_set_home(self):
        towers = parse_dict("samples/towers.json")
        towers = {key: tuple(value) for (key, value) in towers.items()}
        new_home = bc.core.Position(antenna=towers["1"])
        self.user.set_home(new_home)
        self.assertEqual(self.user.home, new_home)

        new_home = bc.core.Position(location=(42.3555368, -71.099507))
        self.user.set_home((42.3555368, -71.099507))
        self.assertEqual(self.user.home, new_home)


class TestTowers(unittest.TestCase):
    def setUp(self):
        self.user = bc.io.read_csv(
            "A", "samples/manual/", "samples/towers.csv", describe=False)

    def testParse(self):
        towers = parse_dict("samples/towers.json")
        towers = {key: tuple(value) for (key, value) in towers.items()}

        self.assertDictEqual(self.user.antennas, towers)


class TestDescribe(unittest.TestCase):
    def setUp(self):
        self.empty_user = bc.User()
        self.sample_user = bc.tests.sample_user()

    def test_describe_empty(self):
        stdout_, stream = sys.stdout, StringIO()

        try:
            sys.stdout = stream
            self.empty_user.describe()
        finally:
            sys.stdout = stdout_

        rv = stream.getvalue().strip()
        baseline = """\x1b[32m[ ]\x1b[0m No records stored
\x1b[32m[ ]\x1b[0m No contacts
\x1b[32m[ ]\x1b[0m No attribute stored
\x1b[32m[ ]\x1b[0m No antenna stored
\x1b[32m[ ]\x1b[0m No recharges
\x1b[32m[ ]\x1b[0m No home
\x1b[32m[ ]\x1b[0m No texts
\x1b[32m[ ]\x1b[0m No calls
\x1b[32m[ ]\x1b[0m No network"""
        self.assertEqual(rv, baseline)

    def test_describe_sample(self):
        stdout_, stream = sys.stdout, StringIO()

        try:
            sys.stdout = stream
            self.sample_user.describe()
        finally:
            sys.stdout = stdout_

        rv = stream.getvalue().strip()
        baseline = """\x1b[32m[x]\x1b[0m 1482 records from 2012-01-01 00:14:24 to 2012-02-27 10:52:45
\x1b[32m[x]\x1b[0m 48 contacts
\x1b[32m[ ]\x1b[0m No attribute stored
\x1b[32m[x]\x1b[0m 7 antennas
\x1b[32m[ ]\x1b[0m No recharges
\x1b[32m[x]\x1b[0m Has home
\x1b[32m[x]\x1b[0m Has texts
\x1b[32m[x]\x1b[0m Has calls
\x1b[32m[x]\x1b[0m Has network"""
        self.assertEqual(rv, baseline)



# The MIT License (MIT)
#
# Copyright (c) 2015-2016 Massachusetts Institute of Technology.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import division

from functools import partial
from datetime import timedelta
import itertools

from bandicoot.helper.maths import mean, std, SummaryStats
from bandicoot.helper.tools import advanced_wrap, AutoVivification, OrderedDict
import numbers


DATE_GROUPERS = {
    None: lambda d: None,
    "day": lambda d: d.isocalendar(),
    "week": lambda d: d.isocalendar()[0:2],
    "month": lambda d: (d.year, d.month),
    "year": lambda d: d.year
}


def filter_user(user, using='records', interaction=None,
                part_of_week='allweek', part_of_day='allday'):
    """
    Filter records of a User objects by interaction, part of week and day.

    Parameters
    ----------
    user : User
        a bandicoot User object
    type : str, default 'records'
        'records' or 'recharges'
    part_of_week : {'allweek', 'weekday', 'weekend'}, default 'allweek'
        * 'weekend': keep only the weekend records
        * 'weekday': keep only the weekdays records
        * 'allweek': use all the records
    part_of_day : {'allday', 'day', 'night'}, default 'allday'
        * 'day': keep only the records during the day
        * 'night': keep only the records during the night
        * 'allday': use all the records
    interaction : object
        The interaction to filter records:
        * "callandtext", for only callandtext;
        * a string, to filter for one type;
        * None, to use all records.
    """

    if using == 'recharges':
        records = user.recharges
    else:
        records = user.records
        if interaction == 'callandtext':
            records = filter(
                lambda r: r.interaction in ['call', 'text'], records)
        elif interaction is not None:
            records = filter(lambda r: r.interaction == interaction, records)

    if part_of_week == 'weekday':
        records = filter(
            lambda r: r.datetime.isoweekday() not in user.weekend, records)
    elif part_of_week == 'weekend':
        records = filter(
            lambda r: r.datetime.isoweekday() in user.weekend, records)
    elif part_of_week != 'allweek':
        raise KeyError(
            "{} is not a valid value for part_of_week. it should be 'weekday', 'weekend' or 'allweek'.".format(part_of_week))

    if user.night_start < user.night_end:
        night_filter = lambda r: user.night_end > r.datetime.time(
        ) > user.night_start
    else:
        night_filter = lambda r: not(
            user.night_end < r.datetime.time() < user.night_start)

    if part_of_day == 'day':
        records = filter(lambda r: not(night_filter(r)), records)
    elif part_of_day == 'night':
        records = filter(night_filter, records)
    elif part_of_day != 'allday':
        raise KeyError(
            "{} is not a valid value for part_of_day. It should be 'day', 'night' or 'allday'.".format(part_of_day))

    return list(records)


def positions_binning(records):
    """
    Bin records by chunks of 30 minutes, returning the most prevalent position.

    If multiple positions have the same number of occurrences
    (during 30 minutes), we select the last one.
    """
    def get_key(d):
        return (d.year, d.day, d.hour, d.minute // 30)

    chunks = itertools.groupby(records, key=lambda r: get_key(r.datetime))

    for _, items in chunks:
        positions = [i.position for i in items]
        # Given the low number of positions per chunk of 30 minutes, and
        # the need for a deterministic value, we use max and not Counter
        yield max(positions, key=positions.count)


def _group_range(records, method):
    """
    Yield the range of all dates between the extrema of
    a list of records, separated by a given time delta.
    """

    start_date = records[0].datetime
    end_date = records[-1].datetime
    _fun = DATE_GROUPERS[method]

    d = start_date

    # Day and week use timedelta
    if method not in ["month", "year"]:
        def increment(i):
            return i + timedelta(**{method + 's': 1})

    elif method == "month":
        def increment(i):
            year, month = divmod(i.month + 1, 12)
            if month == 0:
                month = 12
                year = year - 1
            return d.replace(year=d.year + year, month=month)

    elif method == "year":
        def increment(i):
            return d.replace(year=d.year + 1)

    while _fun(d) <= _fun(end_date):
        yield d
        d = increment(d)


def group_records_with_padding(records, groupby='week'):
    if groupby is None:
        yield records
        return

    if records == []:
        return

    _range = _group_range(records, groupby)
    _fun = DATE_GROUPERS[groupby]

    # Ad hoc grouping with padding
    pointer = next(_range)
    for key, chunk in itertools.groupby(records, key=lambda r: _fun(r.datetime)):
        chunk = list(chunk)

        while _fun(pointer) < key:
            yield []
            pointer = next(_range)

        yield chunk

        pointer = next(_range)


def group_records(records, groupby='week'):
    """
    Group records by year, month, week, or day.

    Parameters
    ----------
    records : iterator
        An iterator over records

    groupby : Default is 'week':
        * 'week': group all records by year and week
        * None: records are not grouped. This is useful if you don't want to
          divide records in chunks
        * "day", "month", and "year" also accepted
    """

    def _group_date(records, _fun):
        for _, chunk in itertools.groupby(records, key=lambda r: _fun(r.datetime)):
            yield list(chunk)

    return _group_date(records, DATE_GROUPERS[groupby])


def infer_type(data):
    """
    Infer the type of objects returned by indicators.

    infer_type returns:
     - 'scalar' for a number or None,
     - 'summarystats' for a SummaryStats object,
     - 'distribution_scalar' for a list of scalars,
     - 'distribution_summarystats' for a list of SummaryStats objects
    """

    if isinstance(data, (type(None), numbers.Number)):
        return 'scalar'

    if isinstance(data, SummaryStats):
        return 'summarystats'

    if hasattr(data, "__len__"):  # list or numpy array
        data = [x for x in data if x is not None]
        if len(data) == 0 or isinstance(data[0], numbers.Number):
            return 'distribution_scalar'
        if isinstance(data[0], SummaryStats):
            return 'distribution_summarystats'

        raise TypeError(
            "{} is not a valid input. It should be a number, a SummaryStats object, or None".format(data[0]))

    raise TypeError(
        "{} is not a valid input. It should be a number, a SummaryStats object, or a list".format(data))


def statistics(data, summary='default', datatype=None):
    """
    Return statistics (mean, standard error, standard error and median, min and max) on data metrics.

    Examples
    --------
    Given a list of integers or floating point numbers,
    ``statistics`` computes the mean and standard error of the mean, and the min and max.

    >>> statistics([0, 1, 2, 3])
    {'mean': 1.5, 'std': 1.2910, 'min': 0, 'max': 3}

    Given a list of ``SummaryStats`` tuples, the function will
    returns the mean, standard error of the mean, min and max for each attribute
    of the tuples.
    """

    def _default_stats(agg):
        if agg is None or len(agg) == 0:
            return OrderedDict([('mean', None), ('std', None)])
        else:
            # Some functions may return None values
            # It's better to filter them
            agg = [x for x in agg if x is not None]
            return OrderedDict([('mean', mean(agg)), ('std', std(agg))])

    def _stats_dict(v):
        rv = [(key, _default_stats([getattr(s, key, None) for s in data])) for key in v]
        return OrderedDict(rv)

    summary_keys = {
        'default': ['mean', 'std'],
        'extended': ['mean', 'std', 'median', 'skewness', 'kurtosis', 'min', 'max']
    }

    if datatype is None:
        datatype = infer_type(data)

    if datatype == 'scalar':
        return data

    if datatype == 'summarystats':
        if summary is None:
            return data.distribution
        elif summary in ['default', 'extended']:
            rv = [(key, getattr(data, key, None)) for key in summary_keys[summary]]
            return OrderedDict(rv)
        else:
            raise ValueError("{} is not a valid summary type".format(summary))

    if datatype == 'distribution_scalar':
        if summary == 'default':
            return _default_stats(data)
        elif summary is None:
            return data
        else:
            raise ValueError("{} is not a valid summary type".format(summary))

    if datatype == 'distribution_summarystats':
        if summary is None:
            return [item.distribution for item in data]
        elif summary in ['extended', 'default']:
            return _stats_dict(summary_keys[summary])
        else:
            raise ValueError("{} is not a valid summary type".format(summary))

    raise ValueError("{} is not a valid data type.".format(datatype))


def _ordereddict_product(dicts):
    return [OrderedDict(zip(dicts, x)) for x in
            itertools.product(*dicts.values())]


def grouping_query(user, query):
    # Filter records for all possible combinations of parameters
    combinations = _ordereddict_product(query['divide_by'])
    params_groups = [(p, filter_user(user, using=query['using'], **p))
                     for p in combinations]

    # Group records by week, month, etc.
    if query['filter_empty']:
        agg_function = group_records
    else:
        agg_function = group_records_with_padding

    def select_function(g):
        if query['binning'] is True:
            return [list(positions_binning(r)) for r in g]
        else:
            return [r for r in g]

    groupby = query['groupby']
    groups = [(p, select_function(agg_function(g, groupby)))
              for p, g in params_groups]

    return groups


def _generic_wrapper(f, user, operations, datatype):
    def compute_indicator(g):
        if operations['apply']['user_kwd']:
            return f(list(g), user, **operations['apply']['kwargs'])
        else:
            return f(list(g), **operations['apply']['kwargs'])

    def map_and_apply(params_combinations):
        for params, groups in params_combinations:
            results = [compute_indicator(g) for g in groups]

            if operations['grouping']['groupby'] is None:
                results = results[0] if len(results) != 0 else None

            stats = statistics(results, datatype=datatype,
                               summary=operations['apply']['summary'])

            yield list(params.values()), stats

    query = operations['grouping']
    groups = user._cached_grouping_query(query)

    returned = AutoVivification()
    for params, stats in map_and_apply(groups):
        returned.insert(params, stats)
    return returned


def divide_parameters(split_week, split_day, interaction):
    if isinstance(interaction, str):
        interaction = [interaction]

    part_of_day = ['allday']
    if split_day:
        part_of_day += ['day', 'night']

    part_of_week = ['allweek']
    if split_week:
        part_of_week += ['weekday', 'weekend']

    if interaction:
        return OrderedDict([
            ('part_of_week', part_of_week),
            ('part_of_day', part_of_day),
            ('interaction', interaction)
        ])
    else:
        return OrderedDict([
            ('part_of_week', part_of_week),
            ('part_of_day', part_of_day)
        ])


def grouping(f=None, interaction=['call', 'text'], summary='default',
             user_kwd=False):
    """
    ``grouping`` is a decorator for indicator functions, used to simplify the
    source code.

    Parameters
    ----------
    f : function
        The function to decorate
    user_kwd : boolean
        If user_kwd is True, the user object will be passed to the decorated
        function
    interaction : 'call', 'text', 'location', or a list
        By default, all indicators use only 'call' and 'text' records, but the
        interaction keywords filters the records passed to the function.
    summary: 'default', 'extended', None
        An indicator returns data statistics, ether *mean* and *std* by
        default, more with 'extended', or the inner distribution with None.
        See :meth:`~bandicoot.helper.group.statistics` for more details.

    See :ref:`new-indicator-label` to learn how to write an indicator with
    this decorator.

    """

    if f is None:
        return partial(grouping, user_kwd=user_kwd, interaction=interaction,
                       summary=summary)

    def wrapper(user, groupby='week', interaction=interaction, summary=summary,
                split_week=False, split_day=False, filter_empty=True,
                datatype=None, **kwargs):

        if interaction is None:
            interaction = ['call', 'text']
        parameters = divide_parameters(split_week, split_day, interaction)

        operations = {
            'grouping': {
                'using': 'records',
                'binning': False,
                'groupby': groupby,
                'filter_empty': filter_empty,
                'divide_by': parameters
            }, 'apply': {
                'user_kwd': user_kwd,
                'summary': summary,
                'kwargs': kwargs
            }
        }

        for i in parameters['interaction']:
            if i not in ['callandtext', 'call', 'text', 'location']:
                raise ValueError("%s is not a valid interaction value. Only "
                                 "'call', 'text', and 'location' are accepted."
                                 % i)

        return _generic_wrapper(f, user, operations, datatype)

    return advanced_wrap(f, wrapper)


def spatial_grouping(f=None, user_kwd=False, summary='default',
                     time_binning=True):
    if f is None:
        return partial(spatial_grouping, user_kwd=user_kwd, summary=summary,
                       time_binning=time_binning)

    def wrapper(user, groupby='week', summary=summary, split_week=False,
                split_day=False, filter_empty=True, datatype=None, **kwargs):

        parameters = divide_parameters(split_week, split_day, None)
        operations = {
            'grouping': {
                'using': 'records',
                'binning': time_binning,
                'groupby': groupby,
                'filter_empty': filter_empty,
                'divide_by': parameters
            }, 'apply': {
                'user_kwd': user_kwd,
                'summary': summary,
                'kwargs': kwargs
            }
        }
        return _generic_wrapper(f, user, operations, datatype)

    return advanced_wrap(f, wrapper)


def recharges_grouping(f=None, summary='default', user_kwd=False):
    if f is None:
        return partial(grouping, user_kwd=user_kwd, summary=summary)

    def wrapper(user, groupby='week', summary=summary,
                split_week=False, split_day=False, filter_empty=True,
                datatype=None, **kwargs):

        parameters = divide_parameters(split_week, split_day, None)

        operations = {
            'grouping': {
                'using': 'recharges',
                'binning': False,
                'groupby': groupby,
                'filter_empty': filter_empty,
                'divide_by': parameters
            }, 'apply': {
                'user_kwd': user_kwd,
                'summary': summary,
                'kwargs': kwargs
            }
        }

        return _generic_wrapper(f, user, operations, datatype)

    return advanced_wrap(f, wrapper)



# The MIT License (MIT)
#
# Copyright (c) 2015-2016 Massachusetts Institute of Technology.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import division

from .helper.group import grouping
from .helper.maths import entropy, summary_stats
from .helper.tools import pairwise
from collections import Counter

import math
import datetime
from collections import defaultdict


@grouping
def interevent_time(records):
    """
    The interevent time between two records of the user.
    """
    inter_events = pairwise(r.datetime for r in records)
    inter = [(new - old).total_seconds() for old, new in inter_events]

    return summary_stats(inter)


@grouping
def number_of_contacts(records, direction=None, more=0):
    """
    The number of contacts the user interacted with.

    Parameters
    ----------
    direction : str, optional
        Filters the records by their direction: ``None`` for all records,
        ``'in'`` for incoming, and ``'out'`` for outgoing.
    more : int, default is 0
        Counts only contacts with more than this number of interactions.
    """
    if direction is None:
        counter = Counter(r.correspondent_id for r in records)
    else:
        counter = Counter(r.correspondent_id for r in records if r.direction == direction)
    return sum(1 for d in counter.values() if d > more)


@grouping
def entropy_of_contacts(records, normalize=False):
    """
    The entropy of the user's contacts.

    Parameters
    ----------
    normalize: boolean, default is False
        Returns a normalized entropy between 0 and 1.

    """
    counter = Counter(r.correspondent_id for r in records)

    raw_entropy = entropy(counter.values())
    n = len(counter)
    if normalize and n > 1:
        return raw_entropy / math.log(n)
    else:
        return raw_entropy


@grouping
def interactions_per_contact(records, direction=None):
    """
    The number of interactions a user had with each of its contacts.

    Parameters
    ----------
    direction : str, optional
        Filters the records by their direction: ``None`` for all records,
        ``'in'`` for incoming, and ``'out'`` for outgoing.
    """
    if direction is None:
        counter = Counter(r.correspondent_id for r in records)
    else:
        counter = Counter(r.correspondent_id for r in records
                          if r.direction == direction)
    return summary_stats(counter.values())


@grouping(user_kwd=True, interaction='call')
def percent_initiated_interactions(records, user):
    """
    The percentage of calls initiated by the user.
    """
    if len(records) == 0:
        return 0

    initiated = sum(1 for r in records if r.direction == 'out')
    return initiated / len(records)


@grouping(user_kwd=True)
def percent_nocturnal(records, user):
    """
    The percentage of interactions the user had at night.

    By default, nights are 7pm-7am. Nightimes can be set in
    ``User.night_start`` and ``User.night_end``.
    """
    if len(records) == 0:
        return 0

    if user.night_start < user.night_end:
        night_filter = lambda d: user.night_end > d.time() > user.night_start
    else:
        night_filter = lambda d: not(user.night_end < d.time() < user.night_start)

    return sum(1 for r in records if night_filter(r.datetime)) / len(records)


@grouping(interaction='call')
def call_duration(records, direction=None):
    """
    The duration of the user's calls.

    Parameters
    ----------
    direction : str, optional
        Filters the records by their direction: ``None`` for all records,
        ``'in'`` for incoming, and ``'out'`` for outgoing.
    """
    if direction is None:
        call_durations = [r.call_duration for r in records]
    else:
        call_durations = [r.call_duration for r in records if r.direction == direction]

    return summary_stats(call_durations)


def _conversations(group, delta=datetime.timedelta(hours=1)):
    """
    Group texts into conversations. The function returns an iterator over records grouped by conversations.

    See :ref:`Using bandicoot <conversations-label>` for a definition of conversations.

    A conversation begins when one person sends a text-message to the other and ends when one of them makes a phone call
    or there is no activity between them for an hour.
    """
    last_time = None
    results = []
    for g in group:

        if last_time is None or g.datetime - last_time < delta:
            if g.interaction == 'text':
                results.append(g)

            # A call always ends a conversation
            else:
                if len(results) != 0:
                    yield results
                    results = []

        else:
            if len(results) != 0:
                yield results

            if g.interaction == 'call':
                results = []
            else:
                results = [g]

        last_time = g.datetime

    if len(results) != 0:
        yield results


@grouping(interaction='callandtext')
def response_rate_text(records):
    """
    The response rate of the user (between 0 and 1).

    Considers text-conversations which began with an incoming text.  Response rate
    is the fraction of such conversations in which the user sent a text (a response).

    The following sequence of messages defines four conversations (``I`` for an
    incoming text, ``O`` for an outgoing text): ::

        I-O-I-O => Started with an incoming text and at least one outgoing text
        I-I-O-I => Started with an incoming text and at least one outgoing text
        I-I-I-I => Started with an incoming text but doesn't have outgoing texts
        O-O-I-O => Not starting with an incoming text

    Here, the ratio would be 2/3 as we have 3 conversations starting with an incoming text and 2 of them have at least one outgoing text.

    See :ref:`Using bandicoot <conversations-label>` for a definition of conversations.
    """
    if len(records) == 0:
        return None

    interactions = defaultdict(list)
    for r in records:
        interactions[r.correspondent_id].append(r)

    def _response_rate(grouped):
        received, responded = 0, 0
        conversations = _conversations(grouped)

        for conv in conversations:
            if len(conv) != 0:
                first = conv[0]
                if first.direction == 'in' and first.interaction == 'text':
                    received += 1
                    if any((i.direction == 'out' for i in conv)):
                        responded += 1

        return responded, received

    # Group all records by their correspondent, and compute the response rate
    # for each
    all_couples = map(_response_rate, list(interactions.values()))
    responded, received = map(sum, list(zip(*all_couples)))

    return responded / received if received != 0 else None


@grouping(interaction='callandtext')
def response_delay_text(records):
    """
    The response delay of the user within a conversation (in seconds)

    The following sequence of messages defines conversations (``I`` for an
    incoming text, ``O`` for an outgoing text, ``-`` for a one minute
    delay): ::

        I-O--I----O, we have a 60 seconds response delay and a 240 seconds response delay
        O--O---I--O, we have a 1200 seconds response delay
        I--II---I-I, we don't have a response delay. The user hasn't answered

    For this user, the distribution of response delays will be ``[60, 240, 60]``

    Notes
    -----
    See :ref:`Using bandicoot <conversations-label>` for a definition of conversations.
    Conversation are defined to be a series of text messages each sent no more than an hour
    after the previous. The response delay can thus not be greater than one hour.
    """
    interactions = defaultdict(list)
    for r in records:
        interactions[r.correspondent_id].append(r)

    def _response_delay(grouped):
        ts = ((b.datetime - a.datetime).total_seconds()
              for conv in _conversations(grouped)
              for a, b in pairwise(conv)
              if b.direction == 'out' and a.direction == 'in')

        return ts

    delays = [r for i in interactions.values() for r in _response_delay(i)
              if r > 0]

    return summary_stats(delays)


@grouping(interaction='callandtext')
def percent_initiated_conversations(records):
    """
    The percentage of conversations that have been initiated by the user.

    Each call and each text conversation is weighted as a single interaction.

    See :ref:`Using bandicoot <conversations-label>` for a definition of
    conversations.
    """
    interactions = defaultdict(list)
    for r in records:
        interactions[r.correspondent_id].append(r)

    def _percent_initiated(grouped):
        mapped = [(1 if conv[0].direction == 'out' else 0, 1)
                  for conv in _conversations(grouped)]
        return mapped

    all_couples = [sublist for i in interactions.values()
                   for sublist in _percent_initiated(i)]

    if len(all_couples) == 0:
        init, total = 0, 0
    else:
        init, total = list(map(sum, list(zip(*all_couples))))

    return init / total if total != 0 else 0


@grouping(interaction='callandtext')
def active_days(records):
    """
    The number of days during which the user was active. A user is considered
    active if he sends a text, receives a text, initiates a call, receives a
    call, or has a mobility point.
    """
    days = set(r.datetime.date() for r in records)
    return len(days)


@grouping
def percent_pareto_interactions(records, percentage=0.8):
    """
    The percentage of user's contacts that account for 80% of its interactions.
    """
    if len(records) == 0:
        return None

    user_count = Counter(r.correspondent_id for r in records)

    target = int(math.ceil(sum(user_count.values()) * percentage))
    user_sort = sorted(user_count.keys(), key=lambda x: user_count[x])

    while target > 0 and len(user_sort) > 0:
        user_id = user_sort.pop()
        target -= user_count[user_id]

    return (len(user_count) - len(user_sort)) / len(records)


@grouping(interaction='call')
def percent_pareto_durations(records, percentage=0.8):
    """
    The percentage of user's contacts that account for 80% of its total time
    spend on the phone. Optionally takes a percentage argument as a decimal
    (e.g., .8 for 80%).
    """
    if len(records) == 0:
        return None

    user_count = defaultdict(int)
    for r in records:
        if r.interaction == "call":
            user_count[r.correspondent_id] += r.call_duration

    target = int(math.ceil(sum(user_count.values()) * percentage))
    user_sort = sorted(user_count.keys(), key=lambda x: user_count[x])

    while target > 0 and len(user_sort) > 0:
        user_id = user_sort.pop()
        target -= user_count[user_id]

    return (len(user_count) - len(user_sort)) / len(records)


@grouping
def balance_of_contacts(records, weighted=True):
    """
    The balance of interactions per contact. For every contact,
    the balance is the number of outgoing interactions divided by the total
    number of interactions (in+out).

    .. math::

       \\forall \\,\\text{contact}\\,c,\\;\\text{balance}\,(c) = \\frac{\\bigl|\\text{outgoing}\,(c)\\bigr|}{\\bigl|\\text{outgoing}\,(c)\\bigr|+\\bigl|\\text{incoming}\,(c)\\bigr|}

    Parameters
    ----------
    weighted : str, optional
        If ``True``, the balance for each contact is weighted by
        the number of interactions the user had with this contact.
    """
    counter_out = defaultdict(int)
    counter = defaultdict(int)

    for r in records:
        if r.direction == 'out':
            counter_out[r.correspondent_id] += 1
        counter[r.correspondent_id] += 1

    if not weighted:
        balance = [counter_out[c] / counter[c] for c in counter]
    else:
        balance = [counter_out[c] / sum(counter.values()) for c in counter]

    return summary_stats(balance)


@grouping()
def number_of_interactions(records, direction=None):
    """
    The number of interactions.

    Parameters
    ----------
    direction : str, optional
        Filters the records by their direction: ``None`` for all records,
        ``'in'`` for incoming, and ``'out'`` for outgoing.
    """
    if direction is None:
        return len(records)
    else:
        return len([r for r in records if r.direction == direction])
